#!/usr/bin/env python3
# -*- coding: utf-8; -*-

import numpy as np

from katoptron.readers import *


def read_iter_all_ensemble_sizes(n_samples, ensemble_sizes, file_base_dir, reduction=False):
    iter = np.zeros((len(ensemble_sizes), n_samples))
    for i in range(0, len(ensemble_sizes)):
        file_base_dir_2 = file_base_dir + '/' + str(ensemble_sizes[i])+'/'
        for j in range(0, int((n_samples // ensemble_sizes[i]))):
            input_MC_file_dir = file_base_dir_2 + str(j)+'/'
            es = ensemble_sizes[i]
            if reduction:
                es = 1
            iter_indices, residuals = read_Belos(
                input_MC_file_dir + 'belos_out.txt', es)
            for k in range(0, ensemble_sizes[i]):
                iter[i][j*ensemble_sizes[i]+k] = iter_indices[-1]
    return iter


def read_timers_all_ensemble_sizes(n_samples, ensemble_sizes, file_base_dir, querylines, use_waves_timers, use_teuchos_timers):
    timers = np.zeros((len(ensemble_sizes), n_samples, len(querylines)))
    timers_s = np.zeros((len(querylines),))
    for i in range(0, len(ensemble_sizes)):
        file_base_dir_2 = file_base_dir + '/' + str(ensemble_sizes[i])+'/'
        for j in range(0, int((n_samples // ensemble_sizes[i]))):
            input_MC_file_dir = file_base_dir_2 + str(j)+'/'
            for k in use_waves_timers:
                timers_s[int(k)] = read_timers(input_MC_file_dir +
                                               'timers.txt', [querylines[int(k)]])[0]
            for k in use_teuchos_timers:
                timers_s[int(k)] = read_teuchos_timers(
                    input_MC_file_dir + 'teuchos_timers.txt', [querylines[int(k)]])[0]
            for k in range(0, ensemble_sizes[i]):
                timers[i][j*ensemble_sizes[i]+k][:] = timers_s[:]

    return timers


def compute_speedup(n_samples, ensemble_sizes, file_base_dir, querylines, per_iteration, quantiles, use_waves_timers, use_teuchos_timers, reduction=False):
    iter = read_iter_all_ensemble_sizes(
        n_samples, ensemble_sizes, file_base_dir, reduction)
    timers = read_timers_all_ensemble_sizes(
        n_samples, ensemble_sizes, file_base_dir, querylines, use_waves_timers, use_teuchos_timers)

    n_quantiles = len(quantiles)

    speedup = np.zeros((len(ensemble_sizes), len(querylines), n_quantiles))

    i_no_ep = 0

    for i in range(0, len(ensemble_sizes)):
        if i == i_no_ep:
            speedup[i][:][:] = np.ones((1, len(querylines), n_quantiles))
        else:
            cum_timers = np.zeros(
                (2, int((n_samples // ensemble_sizes[i])), len(querylines)))
            for j in range(0, int((n_samples // ensemble_sizes[i]))):
                for l in range(0, len(querylines)):
                    if per_iteration[l]:
                        sum_iter = 0
                    else:
                        sum_iter = 1
                    for k in range(0, ensemble_sizes[i]):
                        cum_timers[0][j][l] = cum_timers[0][j][l] + \
                            timers[i_no_ep][j*ensemble_sizes[i]+k][l]
                        if per_iteration[l]:
                            sum_iter = sum_iter + \
                                iter[i_no_ep][j*ensemble_sizes[i]+k]
                    cum_timers[0][j][l] = (
                        cum_timers[0][j][l] / sum_iter)
                    if cum_timers[0][j][l] == 0:
                        print('Warning! The cumulated cpu cost of ' + querylines[l] + ' is zero and is replaced by ' + str(
                            (0.01 / sum_iter)) + ', the precision of the timers has to be increased or the test is too small.')
                        cum_timers[0][j][l] = (0.01 / sum_iter)
                    if per_iteration[l]:
                        cum_timers[1][j][l] = (
                            timers[i][j*ensemble_sizes[i]][l] / (iter[i][j*ensemble_sizes[i]]*ensemble_sizes[i]))
                    else:
                        cum_timers[1][j][l] = timers[i][j*ensemble_sizes[i]][l]
            for l in range(0, len(querylines)):
                speedup[i][l] = np.quantile(
                    (cum_timers[0, :, l] / cum_timers[1, :, l]), quantiles, axis=0)
    return speedup


def plot_speedup(speedup, ensemble_sizes, which_l=[0], title=None, querylines=None, per_iteration=None):
    import matplotlib.pyplot as plt

    plt.figure(figsize=(8, 8))

    col = {'c1': (0., 0.4470, 0.7410),     'c2': (0.8500, 0.3250, 0.0980),
           'c3': (0.9290, 0.6940, 0.1250), 'c4': (0.4940, 0.1840, 0.5560),
           'c5': (0.4660, 0.6740, 0.1880), 'c6': (0.3010, 0.7450, 0.9330),
           'c7': (0.6350, 0.0780, 0.1840), 'c8': (0, 0.4470, 0.7410)}

    plt_linewidth = 2
    plt.style.use('ggplot')

    for i in range(0, len(which_l)):
        color = col['c'+str(i+1)]
        for j in range(0, speedup.shape[2]):
            if j == 0 and querylines is not None:
                plt.plot(ensemble_sizes, speedup[:, which_l[i], j], color=color,
                         label=querylines[which_l[i]]+'_'+str(per_iteration[which_l[i]]))
            else:
                plt.plot(ensemble_sizes,
                         speedup[:, which_l[i], j], color=color)

    plt.xlabel('ensemble size')
    plt.ylabel('speedup')
    plt.xlim(min(ensemble_sizes), max(ensemble_sizes))
    if title is not None:
        plt.title(title)
    if querylines is not None:
        plt.legend()


def save_speedup(speedup, ensemble_sizes, querylines, filename):
    tmp = np.zeros((speedup.shape[0], speedup.shape[2]+1))
    tmp[:, 0] = ensemble_sizes
    for l in range(0, len(querylines)):
        for i in range(1, speedup.shape[2]+1):
            tmp[:, i] = speedup[:, l, i-1]
        name = filename + '_' + querylines[l] + '.txt'
        name = name.replace(" ", "_")
        name = name.replace(":", "")
        name = name.replace("[", "")
        name = name.replace("]", "")
        name = name.replace("(", "")
        name = name.replace(")", "")
        name = name.replace("*", "")
        np.savetxt(name, tmp, delimiter=' ')


def compute_R(n_samples, ensemble_sizes, file_base_dir, quantiles, reduction=False):
    iter = read_iter_all_ensemble_sizes(
        n_samples, ensemble_sizes, file_base_dir, reduction)
    n_quantiles = len(quantiles)

    R = np.ones((len(ensemble_sizes), n_quantiles))

    i_no_ep = 0

    for i in range(0, len(ensemble_sizes)):
        if i != i_no_ep:
            cum_iter = np.zeros(
                (2, int((n_samples // ensemble_sizes[i]))))
            for j in range(0, int((n_samples // ensemble_sizes[i]))):
                for k in range(0, ensemble_sizes[i]):
                    cum_iter[0][j] = cum_iter[0][j] + \
                        iter[i_no_ep][j*ensemble_sizes[i]+k]
                cum_iter[1][j] = iter[i][j*ensemble_sizes[i]]*ensemble_sizes[i]
            R[i, :] = np.quantile(
                (cum_iter[1, :] / cum_iter[0, :]), quantiles, axis=0)
    return R


def save_R(R, filename):
    name = filename + '.txt'
    name = name.replace(" ", "_")
    name = name.replace(":", "")
    name = name.replace("[", "")
    name = name.replace("]", "")
    name = name.replace("(", "")
    name = name.replace(")", "")
    name = name.replace("*", "")
    np.savetxt(name, R, delimiter=' ')
