#ifndef KATOPTRON_SPECTRALAPPROACH_H
#define KATOPTRON_SPECTRALAPPROACH_H

#include "EnsembleTraits.h"
#include <cmath>
#include <random>

/**
 * @brief A function which computes a random matrix \f$\boldsymbol{Z}\f$ starting at a given seed.
 * 
 * This matrix is used to generate a random field based on the spectral approach
 */
template <typename Scalar, typename Device>
Kokkos::View<Scalar **, Kokkos::LayoutLeft, Device> construct_Z(size_t muw, unsigned long seed_Z, size_t wait = 0)
{
    Kokkos::View<Scalar **, Kokkos::LayoutLeft, Device> Z = Kokkos::View<Scalar **, Kokkos::LayoutLeft, Device>("Z", muw, muw);

    std::default_random_engine generator;
    generator.seed(seed_Z);
    std::uniform_real_distribution<double> distribution(0.0, 1.0);
    // Redraw the data related to the previous samples to start at the good place
    for (size_t i = 0; i < wait; ++i)
        distribution(generator);

    for (size_t ell = 0; ell < EnsembleTraits<Scalar>::size; ++ell)
        for (size_t i = 0; i < muw; ++i)
            for (size_t j = 0; j < muw; ++j)
                EnsembleTraits<Scalar>::coeff(Z(i, j), ell) = sqrt(-log(distribution(generator)));

    return Z;
}

/**
 * @brief A function which computes a random matrix \f$\boldsymbol{\Phi}\f$ starting at a given seed.
 * 
 * This matrix is used to generate a random field based on the spectral approach
 */
template <typename Scalar, typename Device>
Kokkos::View<Scalar **, Kokkos::LayoutLeft, Device> construct_Phi(size_t muw, unsigned long seed_Phi, size_t wait = 0)
{
    Kokkos::View<Scalar **, Kokkos::LayoutLeft, Device> Phi = Kokkos::View<Scalar **, Kokkos::LayoutLeft, Device>("Phi", muw, muw);

    std::default_random_engine generator;
    generator.seed(seed_Phi);
    std::uniform_real_distribution<double> distribution(0.0, 1.0);
    // Redraw the data related to the previous samples to start at the good place
    for (size_t i = 0; i < wait; ++i)
        distribution(generator);

    for (size_t ell = 0; ell < EnsembleTraits<Scalar>::size; ++ell)
        for (size_t i = 0; i < muw; ++i)
            for (size_t j = 0; j < muw; ++j)
                EnsembleTraits<Scalar>::coeff(Phi(i, j), ell) = 2. * M_PI * distribution(generator);

    return Phi;
}

/**
 * @brief A class that constructs realization(s) of Gaussian or exponential
 * scalar random field based on the spectral approach.
 */
template <typename Scalar, typename Device>
class SpectralApproach
{
public:
    typedef typename Teuchos::ScalarTraits<Scalar>::coordinateType MeshScalar;

    Kokkos::View<Scalar **, Kokkos::LayoutLeft, Device> Z;   // Z=sqrt(-log(rand(muw,muw)));
    Kokkos::View<Scalar **, Kokkos::LayoutLeft, Device> Phi; // Phi=2*pi*rand(muw,muw);
    Kokkos::View<MeshScalar *, Kokkos::LayoutLeft, Device> w;
    Kokkos::View<MeshScalar *, Kokkos::LayoutLeft, Device> S;

    MeshScalar a, r, Dw;
    size_t muw;

    SpectralApproach(Kokkos::View<Scalar **, Kokkos::LayoutLeft, Device> _Z,
                     Kokkos::View<Scalar **, Kokkos::LayoutLeft, Device> _Phi,
                     MeshScalar _a = 20, MeshScalar _r = 8, size_t _muw = 512,
                     bool Gaussian = true) : Z(_Z), Phi(_Phi), a(_a), r(_r), muw(_muw)
    {
        MeshScalar wL = M_PI / a * r;
        Dw = 2 * wL / muw;

        w = Kokkos::View<MeshScalar *, Kokkos::LayoutLeft, Device>("w", muw);
        S = Kokkos::View<MeshScalar *, Kokkos::LayoutLeft, Device>("w", muw);

        for (size_t i = 0; i < muw; ++i)
            w(i) = -wL + (i + 0.5) * Dw;

        if (Gaussian)
            for (size_t i = 0; i < muw; ++i)
                S(i) = a / sqrt(2. * M_PI) * exp(-pow(a * w(i), 2) / 2.);
        else
            for (size_t i = 0; i < muw; ++i)
                S(i) = a / (M_PI * (1 + pow(a, 2) * pow(w(i), 2)));
    }

    KOKKOS_INLINE_FUNCTION
    Scalar operator()(const MeshScalar x, const MeshScalar y) const
    {
        //Teuchos::TimeMonitor LocalTimer (*Random_field_eval);

        Scalar output = 0.;

        for (size_t i = 0; i < muw; ++i)
            for (size_t j = 0; j < muw; ++j)
                output += sqrt(S(i) * S(j)) * Z(i, j) * cos(Phi(i, j) + w(i) * x + w(j) * y);

        output *= sqrt(2) * Dw;

        return output;
    }
};

#endif //KATOPTRON_SPECTRALAPPROACH_H
