#!/usr/bin/env python3
# -*- coding: utf-8; -*-

import katoptron as m
import tbox
import tbox.gmsh as gmsh
import katoptron.utilities as u
import katoptron.eigenvalues as e
import katoptron.readers as r
import katoptron.convergence as c
from PyTrilinos import Teuchos
import shlex
import subprocess
import os
import numpy as np
import tarfile
import glob
from fwk.wutils import parseargs
import matplotlib.pyplot as plt


def read_mm_vector(filename, ensemble_size):
    with open(filename, 'r') as my_file:
        text = my_file.read().splitlines(True)
        if ensemble_size > 1:
            for i in range(2, len(text)):
                text[i] = text[i].replace("[", "")
                text[i] = text[i].replace("]", "")

    X = np.genfromtxt(text[2:])
    return X


def read_mm_sparse_matrix(filename, ensemble_size):
    with open(filename, 'r') as my_file:
        text = my_file.read().splitlines(True)
        if ensemble_size > 1:
            for i in range(2, len(text)):
                text[i] = text[i].replace("[", "")
                text[i] = text[i].replace("]", "")

    sizes = [375, 375]  # np.genfromtxt(text[1])
    data = np.genfromtxt(text[2:])
    A = np.zeros((int(sizes[0]), int(sizes[1]), ensemble_size))
    for i in range(0, data.shape[0]):
        A[int(data[i, 0]) - 1, int(data[i, 1]) - 1, :] = data[i, 2:]
    return A


def extract_values(filename, node_id, which_dof, dof_per_node, ensemble_size):
    output = np.zeros((ensemble_size, ))
    X = read_mm_vector(filename, ensemble_size)
    output = X[dof_per_node * node_id + which_dof, :]
    return output


def compare_QoI(base_dir_NER, base_dir_ER, case_ids, ensemble_sizes):
    node_id = 46
    which_dof = 2
    dof_per_node = 3

    for case_id in case_ids:
        for ensemble_size in ensemble_sizes:
            if case_id == 2:
                param_min = 0.075
                param_max = 0.125
            elif case_id == 6:
                param_min = 0.95
                param_max = 1.05
            elif case_id == 7:
                param_min = 1.2
                param_max = 1.3
            else:
                param_min = 0.75
                param_max = 1.25

            param = np.linspace(param_min, param_max, ensemble_size)
            solution_file_name = base_dir_NER + 'test_' + str(
                case_id) + '_s_' + str(ensemble_size) + '/x_mm.txt'
            output_NER = extract_values(solution_file_name, node_id, which_dof,
                                        dof_per_node, ensemble_size)
            solution_file_name = base_dir_ER + 'test_' + str(
                case_id) + '_s_' + str(ensemble_size) + '/x_mm.txt'
            output_ER = extract_values(solution_file_name, node_id, which_dof,
                                       dof_per_node, ensemble_size)
            plt.figure()
            plt.plot(output_NER, 'o-')
            plt.plot(output_ER, 'o-')
            plt.savefig('QoI_case_' + str(case_id) + '_s_' +
                        str(ensemble_size) + '.pdf')

            f = open(
                'QoI_case_' + str(case_id) + '_s_' + str(ensemble_size) +
                '.txt', 'w')
            for l in range(0, ensemble_size):
                print(str(param[l]) + ' ' + str(
                    output_NER[l]) + ' ' + str(
                        output_ER[l]) + ' ' + str(output_NER[l] - output_ER[l]), file=f)
            f.closed


def read_Belos(base_dir_NER, base_dir_ER, case_ids, ensemble_sizes):
    for case_id in case_ids:
        for ensemble_size in ensemble_sizes:
            filename_NER = base_dir_NER + 'test_' + str(case_id) + '_s_' + str(
                ensemble_size) + '/belos_out.txt'
            iter_indices_NER, residuals_NER = r.read_Belos(
                filename_NER, ensemble_size)

            filename_ER = base_dir_ER + 'test_' + str(case_id) + '_s_' + str(
                ensemble_size) + '/belos_out.txt'
            iter_indices_ER, residuals_ER = r.read_Belos(filename_ER, 1)

            f = open(
                'Belos_NER_' + str(case_id) + '_s_' + str(ensemble_size) +
                '.txt', 'w')
            for i in range(0, len(iter_indices_NER)):
                tmp = str(i) + ' '
                for l in range(0, ensemble_size):
                    tmp = tmp + str(residuals_NER[i, l]) + ' '
                print(tmp, file=f)
            f.closed

            f = open(
                'Belos_ER_' + str(case_id) + '_s_' + str(ensemble_size) +
                '.txt', 'w')
            for i in range(0, len(iter_indices_ER)):
                print(str(i) + ' ' + str(residuals_ER[i]), file=f)
            f.closed

            plt.figure()
            plt.semilogy(iter_indices_NER, residuals_NER[:, 0], 'o-')
            plt.semilogy(iter_indices_ER, residuals_ER, 'o-')
            plt.savefig('Belos_ER_' + str(case_id) + '_s_' +
                        str(ensemble_size) + '.pdf')


def compute_cum_delta_normal(lambdas,
                             omegas,
                             lambda_min=None,
                             lambda_max=None,
                             n=200):

    indices = np.argsort(lambdas)
    lambdas = lambdas[indices]
    omegas = omegas[indices]

    if lambda_min is None:
        lambda_min = lambdas[0]
    if lambda_max is None:
        lambda_max = lambdas[-1]

    lambda_discrete = np.linspace(lambda_min, lambda_max, n)
    dlambda = ((lambda_max - lambda_min) / (n - 1))

    cum_delta = np.zeros((n, ))
    deriv_delta = np.zeros((n, ))
    cum_delta_omega = np.zeros((n, ))
    deriv_delta_omega = np.zeros((n, ))

    n_lambdas = len(lambdas)

    first_index = 0
    second_index = 0
    for i in range(1, n):
        for j in range(second_index, n_lambdas):
            if lambda_discrete[i] < lambdas[j]:
                second_index = j
                break
        cum_delta[i] = cum_delta[i - 1] + second_index - first_index
        cum_delta_omega[i] = cum_delta_omega[i - 1] + np.sum(
            omegas[first_index:second_index])
        deriv_delta[i] = ((second_index - first_index) / dlambda)
        deriv_delta_omega[i] = (np.sum(
            omegas[first_index:second_index]) / dlambda)
        first_index = second_index

    percent_delta = (cum_delta / cum_delta[n - 1])
    percent_delta_omega = (cum_delta_omega / cum_delta_omega[n - 1])

    return lambda_discrete, cum_delta, deriv_delta, cum_delta_omega, deriv_delta_omega, percent_delta, percent_delta_omega


def compute_deltas_omegas(base_dir, case_ids, ensemble_sizes):
    for case_id in case_ids:
        for ensemble_size in ensemble_sizes:

            filename_b = base_dir + 'test_' + str(case_id) + '_s_' + str(
                ensemble_size) + '/b_mm_after_bc.txt'
            filename_A = base_dir + 'test_' + str(case_id) + '_s_' + str(
                ensemble_size) + '/A_mm_after_bc.txt'

            b = read_mm_vector(filename_b, ensemble_size)
            A = read_mm_sparse_matrix(filename_A, ensemble_size)

            for l in range(0, ensemble_size):
                b_l = b[:, l]
                A_l = e.get_symmetrical_part(A[:, :, l])
                deltas, omegas = e.compute_deltas_omegas_normal(A_l, b_l)
                indices = np.argsort(np.real(deltas))
                f = open(
                    'deltas_omegas_' + str(case_id) + '_s_' +
                    str(ensemble_size) + '_l_' + str(l) + '.txt', 'w')
                for i in range(0, len(deltas)):
                    print(str(np.real(deltas[indices[i]])) + ' ' + str(
                        np.imag(deltas[indices[i]])) + ' ' + str(
                            omegas[indices[i]]) + ' ' + str(i), file=f)
                f.closed

                lambda_discrete, cum_delta, deriv_delta, cum_delta_omega, deriv_delta_omega, percent_delta, percent_delta_omega = compute_cum_delta_normal(
                    np.real(deltas),
                    omegas,
                    lambda_min=0.,
                    lambda_max=12.,
                    n=300)

                plt.figure()
                plt.plot(lambda_discrete, cum_delta)
                plt.savefig('cum_delta_' + str(case_id) + '_s_' +
                            str(ensemble_size) + '_l_' + str(l) + '.pdf')
                plt.close()
                plt.figure()
                plt.plot(lambda_discrete, deriv_delta)
                plt.savefig('deriv_delta_' + str(case_id) + '_s_' +
                            str(ensemble_size) + '_l_' + str(l) + '.pdf')
                plt.close()
                plt.figure()
                plt.plot(lambda_discrete, cum_delta_omega)
                plt.savefig('cum_delta_omega_' + str(case_id) + '_s_' +
                            str(ensemble_size) + '_l_' + str(l) + '.pdf')
                plt.close()
                plt.figure()
                plt.plot(lambda_discrete, deriv_delta_omega)
                plt.savefig('deriv_delta_omega_' + str(case_id) + '_s_' +
                            str(ensemble_size) + '_l_' + str(l) + '.pdf')
                plt.close()

                if l == 0:
                    deltas_total = np.real(deltas)
                    omegas_total = omegas
                else:
                    deltas_total = np.append(deltas_total, np.real(deltas))
                    omegas_total = np.append(omegas_total, omegas)

            lambda_discrete, cum_delta, deriv_delta, cum_delta_omega, deriv_delta_omega, percent_delta, percent_delta_omega = compute_cum_delta_normal(
                deltas_total,
                omegas_total,
                lambda_min=0.,
                lambda_max=12.,
                n=300)

            plt.figure()
            plt.plot(lambda_discrete, cum_delta)
            plt.savefig('cum_delta_' + str(case_id) + '_s_' +
                        str(ensemble_size) + '_all.pdf')
            plt.close()
            plt.figure()
            plt.plot(lambda_discrete, deriv_delta)
            plt.savefig('deriv_delta_' + str(case_id) + '_s_' +
                        str(ensemble_size) + '_all.pdf')
            plt.close()
            plt.figure()
            plt.plot(lambda_discrete, cum_delta_omega)
            plt.savefig('cum_delta_omega_' + str(case_id) + '_s_' +
                        str(ensemble_size) + '_all.pdf')
            plt.close()
            plt.figure()
            plt.plot(lambda_discrete, deriv_delta_omega)
            plt.savefig('deriv_delta_omega_' + str(case_id) + '_s_' +
                        str(ensemble_size) + '_all.pdf')
            plt.close()
            '''
            n_per_l = len(b_l)
            n_total = n_per_l*ensemble_size
            b_total = np.zeros((n_total,))
            A_total = np.zeros((n_total,n_total))
            for l in range(0,ensemble_size):
                b_total[n_per_l*l:n_per_l*(l+1)] = b[:,l]
                A_total[n_per_l*l:n_per_l*(l+1),n_per_l*l:n_per_l*(l+1)] = e.get_symmetrical_part(A[:,:,l])
            deltas_total, omegas_total = e.compute_deltas_omegas_normal(A_total,b_total)

            lambda_discrete, cum_delta, deriv_delta, cum_delta_omega, deriv_delta_omega, percent_delta, percent_delta_omega = compute_cum_delta_normal(deltas_total,omegas_total, lambda_min=0., lambda_max=12., n = 300  )            

            plt.figure()
            plt.plot(lambda_discrete, cum_delta)
            plt.savefig('cum_delta_' + str(case_id) + '_s_' + str(ensemble_size) + '_all_2.pdf')
            plt.close()
            plt.figure()
            plt.plot(lambda_discrete, deriv_delta)
            plt.savefig('deriv_delta_' + str(case_id) + '_s_' + str(ensemble_size) + '_all_2.pdf')             
            plt.close()
            plt.figure()
            plt.plot(lambda_discrete, cum_delta_omega)
            plt.savefig('cum_delta_omega_' + str(case_id) + '_s_' + str(ensemble_size) + '_all_2.pdf')                                
            plt.close()
            plt.figure()
            plt.plot(lambda_discrete, deriv_delta_omega)
            plt.savefig('deriv_delta_omega_' + str(case_id) + '_s_' + str(ensemble_size) + '_all_2.pdf')             
            plt.close() 
            '''


def compute_convergence(base_dir, case_ids, ensemble_sizes, k_max):
    for case_id in case_ids:
        for ensemble_size in ensemble_sizes:
            for l in range(0, ensemble_size):
                data = np.genfromtxt('deltas_omegas_' + str(case_id) + '_s_' +
                                     str(ensemble_size) + '_l_' + str(l) +
                                     '.txt')
                deltas = data[:, 0]
                omegas = data[:, 2]

                n = len(deltas)

                convergence = c.compute_convergence_serial_normal(k_max,
                                                                  n,
                                                                  omegas,
                                                                  deltas,
                                                                  verbose=True)


def main():
    base_dir_NER = '/Volumes/HD_SONY/Thesis/Test_1/all_results_NER_thesis1/'
    base_dir_ER = '/Volumes/HD_SONY/Thesis/Test_1/all_results_ER_thesis1/'

    case_ids = [1, 2, 3, 4, 5, 6, 7]
    ensemble_sizes = [8, 16, 32]

    compare_QoI(base_dir_NER, base_dir_ER, case_ids, ensemble_sizes)
    read_Belos(base_dir_NER, base_dir_ER, case_ids, ensemble_sizes)
    #compute_deltas_omegas(base_dir_NER, case_ids, ensemble_sizes)

    k_max = 3
    case_ids = [1]
    ensemble_sizes = [8]

    #compute_convergence(base_dir_NER, case_ids, ensemble_sizes, k_max)


if __name__ == "__main__":
    main()
